;; -*- mode: scheme; coding: utf-8 -*-
;; Copyright (c) 2025 Vadym Kochan <vadim4j@gmail.com>
;; SPDX-License-Identifier: MIT
#!r6rs

(library (onie info)
  (export read-onie-info
          write-onie-info

          make-onie-info
          onie-info?

          onie-info-product-name
          onie-info-product-name-set!
          onie-info-part-number
          onie-info-part-number-set!
          onie-info-serial-number
          onie-info-serial-number-set!
          onie-info-mac-base
          onie-info-mac-base-set!
          onie-info-manufacture-date
          onie-info-manufacture-date-set!
          onie-info-device-version
          onie-info-device-version-set!
          onie-info-label-revision
          onie-info-label-revision-set!
          onie-info-platform-name
          onie-info-platform-name-set!
          onie-info-onie-version
          onie-info-onie-version-set!
          onie-info-num-macs
          onie-info-num-macs-set!
          onie-info-manufacturer
          onie-info-manufacturer-set!
          onie-info-country-code
          onie-info-country-code-set!
          onie-info-vendor
          onie-info-vendor-set!
          onie-info-diag-version
          onie-info-diag-version-set!
          onie-info-service-tag
          onie-info-service-tag-set!
          onie-info-vendor-extension
          onie-info-vendor-extension-set!
          onie-info-crc
          onie-info-crc-valid?)
  (import (rnrs)
          (hashing crc))

(define-record-type onie-info
  (fields (mutable version)
          (mutable size)
          (mutable product-name)
          (mutable part-number)
          (mutable serial-number)
          (mutable mac-base)
          (mutable manufacture-date)
          (mutable device-version)
          (mutable label-revision)
          (mutable platform-name)
          (mutable onie-version)
          (mutable num-macs)
          (mutable manufacturer)
          (mutable country-code)
          (mutable vendor)
          (mutable diag-version)
          (mutable service-tag)
          (mutable vendor-extension)
          (mutable crc)
          (mutable crc-valid?))
  (protocol
    (lambda (new)
      (lambda ()
        (new #f 0 #f #f #f #f #f #f #f #f #f #f
             #f #f #f #f #f #f #f #f)))))

(define ascii-encoder (make-transcoder (latin-1-codec)))
(define onie-info-sig "TlvInfo\x0;")
(define-crc crc-32)

(define (read-onie-info port)
  (let ([sig (get-bytevector-n port 8)])
    (when (not (equal? (utf8->string sig) onie-info-sig))
      (error 'read-onie-info "Invalid ONIE info signature" sig))
    (let* ([info (make-onie-info)]
           [ver (get-u8 port)]
           [size-bv (get-bytevector-n port 2)]
           [size (bytevector-u16-ref size-bv 0 (endianness big))])
      (onie-info-version-set! info ver)
      (onie-info-size-set! info size)
      (let loop ([size size] [crc (crc-32-update (crc-32-update (crc-32-update (crc-32-init) sig)
                                                                (make-bytevector 1 ver))
                                                 size-bv)])
        (when (positive? size)
          (let* ([type/len-bv (get-bytevector-n port 2)]
                 [type (bytevector-u8-ref type/len-bv 0)]
                 [len (bytevector-u8-ref type/len-bv 1)]
                 [val (get-bytevector-n port len)])
            (case type
              ((#x21)
               (onie-info-product-name-set! info (bytevector->string val ascii-encoder)))
              ((#x22)
               (onie-info-part-number-set! info (bytevector->string val ascii-encoder)))
              ((#x23)
               (onie-info-serial-number-set! info (bytevector->string val ascii-encoder)))
              ((#x24)
               (when (not (eqv? len 6))
                 (error 'read-onie-info "Invalid Base MAC address length (expected 6)" len))
               (onie-info-mac-base-set! info val))
              ((#x25)
               (when (not (eqv? len 19))
                 (error 'read-onie-info "Invalid Manufacture Date length (expected 19)" len))
               (onie-info-manufacture-date-set! info (bytevector->string val ascii-encoder)))
              ((#x26)
               (when (not (eqv? len 1))
                 (error 'read-onie-info "Invalid Device Version length (expected 1)" len))
               (onie-info-device-version-set! info (bytevector-u8-ref val 0)))
              ((#x27)
               (onie-info-label-revision-set! info (bytevector->string val ascii-encoder)))
              ((#x28)
               (onie-info-platform-name-set! info (bytevector->string val ascii-encoder)))
              ((#x29)
               (onie-info-onie-version-set! info (bytevector->string val ascii-encoder)))
              ((#x2a)
               (when (not (eqv? len 2))
                 (error 'read-onie-info "Invalid Num MACs length (expected 2)" len))
               (onie-info-num-macs-set! info (bytevector-u16-ref val 0 (endianness big))))
              ((#x2b)
               (onie-info-manufacturer-set! info (bytevector->string val ascii-encoder)))
              ((#x2c)
               (when (not (eqv? len 2))
                 (error 'read-onie-info "Invalid Country Code length (expected 2)" len))
               (onie-info-country-code-set! info (bytevector->string val ascii-encoder)))
              ((#x2d)
               (onie-info-vendor-set! info (bytevector->string val ascii-encoder)))
              ((#x2e)
               (onie-info-diag-version-set! info (bytevector->string val ascii-encoder)))
              ((#x2f)
               (onie-info-service-tag-set! info (bytevector->string val ascii-encoder)))
              ((#xfd)
               (onie-info-vendor-extension-set! info val))
              ((#xfe)
               (let ([crc-orig (bytevector-u32-ref val 0 (endianness big))]
                     [crc-calc (crc-32-finish (crc-32-update crc type/len-bv))])
                 (onie-info-crc-valid?-set! info (eqv? crc-orig crc-calc))
                 (onie-info-crc-set! info crc-orig)))
            )
	    ;; crc32 tlv must be a last one
            (if (eqv? type #xfe)
                (loop 0 #f)
                (loop (- size len 2) (crc-32-update (crc-32-update crc type/len-bv) val))))))
      info)))

(define (write-onie-info info port)
  (define (put-tlv outp type len val)
    (let* ([val (if (string? val)
                    (string->bytevector val ascii-encoder)
                    val)]
           [len (if (bytevector? val)
                    (bytevector-length val)
                    len)])
      (put-u8 outp type)
      (put-u8 outp len)
      (when (and val (positive? len))
        (cond
          ((and (fixnum? val) (eqv? len 1))
           (put-u8 outp val))
          ((and (fixnum? val) (eqv? len 2))
           (let ([bv (make-bytevector 2)])
             (bytevector-u16-set! bv 0 val (endianness big))
             (put-bytevector outp bv)))
          ((and (fixnum? val) (eqv? len 4))
           (let ([bv (make-bytevector 4)])
             (bytevector-u32-set! bv 0 val (endianness big))
             (put-bytevector outp bv)))
          (else
           (put-bytevector outp val))))))

  (define (fill-info-bytes)
    (let-values ([(outp get-tlv-bytes) (open-bytevector-output-port)])
      ;; add empty header
      (put-bytevector outp (make-bytevector 11 0))
      (when (onie-info-product-name info)
        (put-tlv outp #x21 #f (onie-info-product-name info)))
      (when (onie-info-part-number info)
        (put-tlv outp #x22 #f (onie-info-part-number info)))
      (when (onie-info-serial-number info)
        (put-tlv outp #x23 #f (onie-info-serial-number info)))
      (when (onie-info-mac-base info)
        (put-tlv outp #x24 #f (onie-info-mac-base info)))
      (when (onie-info-manufacture-date info)
        (put-tlv outp #x25 #f (onie-info-manufacture-date info)))
      (when (onie-info-device-version info)
        (put-tlv outp #x26 1 (onie-info-device-version info)))
      (when (onie-info-label-revision info)
        (put-tlv outp #x27 #f (onie-info-label-revision info)))
      (when (onie-info-platform-name info)
        (put-tlv outp #x28 #f (onie-info-platform-name info)))
      (when (onie-info-onie-version info)
        (put-tlv outp #x29 #f (onie-info-onie-version info)))
      (when (onie-info-num-macs info)
        (put-tlv outp #x2a 2 (onie-info-num-macs info)))
      (when (onie-info-manufacturer info)
        (put-tlv outp #x2b #f (onie-info-manufacturer info)))
      (when (onie-info-country-code info)
        (put-tlv outp #x2c #f (onie-info-country-code info)))
      (when (onie-info-vendor info)
        (put-tlv outp #x2d #f (onie-info-vendor info)))
      (when (onie-info-diag-version info)
        (put-tlv outp #x2e #f (onie-info-diag-version info)))
      (when (onie-info-service-tag info)
        (put-tlv outp #x2f #f (onie-info-service-tag info)))
      (when (onie-info-vendor-extension info)
        (put-tlv outp #xfd #f (onie-info-vendor-extension info)))
      ;; CRC-32 TLV value will be appended later
      (put-tlv outp #xfe 4 #f)
      (get-tlv-bytes)))

  (let ([bytes (fill-info-bytes)])
    ;; "TlvInfo"
    (bytevector-copy! (string->bytevector onie-info-sig ascii-encoder) 0 bytes 0 8)
    ;; ONIE version (default 1)
    (bytevector-u8-set! bytes 8 1)
    ;; Size of the TLV data (w/o 11 bytes of the header)
    (bytevector-u16-set! bytes 9 (fx- (bytevector-length bytes) 11) (endianness big))
    (put-bytevector port bytes)
    ;; CRC-32 is the last TLV value entry
    (let ([crc (make-bytevector 4 0)])
      (bytevector-u32-set! crc 0 (crc-32 bytes) (endianness big))
      (put-bytevector port crc))))

)
